@@ -1,16 +1,29 @@
 package CGI::apacheSSI;
 use strict;
+use warnings;
+
+# CHANGES for 0.95:
+
+# DONE:
+# include virtual should not be making http requests to other servers.. 
+# include virtual can be absolute.. 
+# include file cannot be absolute.. 
+# flastmod, fsize
+
+# TODO:
+# move up the  %allowed_tag_count;
+# $allowed_tag_count{'if'}=["expr"]; should use arrays: @allowed_tag_count{'if'}=("expr");
+# handle encoding in echo()
+# PROPER VIRTUAL CALLS TO CGI SCRIPTS AND mod_rewrite URLS
+
 
-# use HTML::SimpleParse;
 use File::Spec::Functions; # catfile()
-use FindBin;
-use LWP::UserAgent;
 use HTTP::Response;
 use HTTP::Cookies;
-use URI;
 use Date::Format;
+use Cwd;
 
-our $VERSION = '0.94';
+our $VERSION = '0.95';
 
 our $DEBUG = 0;
 
@@ -25,7 +38,7 @@ sub import {
 
 my($gmt,$loc,$lmod);
 
-			# NOTE: check for escaped \( or \), what should it do?
+			# NOTE: check for escaped \( or \), what should it do? -- DONE?
 our $L; # used to return the brackets count
 our $RE_parens_2C = qr/
       (	  # g1, everything inside the brackets, incl brackets
@@ -73,7 +86,7 @@ our $RE_apache_expr_quote	  = qr/ (?:"(?:[^"\\]|[\\]+[^\\])*?")
 # NOTE: quotes that would be openers which are immediately preceeded by \w are treated as \w
 	  # NOTE: needs to be preceeded by \s or =, otherwise becomes part of token (parsing oddity with apache 2.2.22)
 our $RE_apache_expr_quote_all	  = qr/  $RE_apache_expr_quote | [^'"`\s]/x;
-our $RE_runaway = qr/ \s+  \w+['"`]\S*\s+[^'"`]+['"`]+  /x;
+our $RE_runaway   =  qr/ \s+  \w+['"`]\S*\s+[^'"`]+['"`]+  /x;
 our $RE_token_NC  =  qr{[[:alpha:]]\S+? (?:\s+ $RE_apache_expr_quote_all*? )*?   $RE_runaway? }x; 
 
 
@@ -91,23 +104,28 @@ sub new {
     tie $loc, 'CGI::apacheSSI::Local', $self;
     tie $lmod, 'CGI::apacheSSI::LMOD', $self;
 
-    $ENV{'DOCUMENT_ROOT'} ||= '';
+    # $ENV{'DOCUMENT_ROOT'} ||= '';
     $self->{'_variables'}     = {
         DOCUMENT_URI    =>  ($args{'DOCUMENT_URI'} || $ENV{'SCRIPT_NAME'}),
+        DOCUMENT_NAME   =>  ($args{'DOCUMENT_NAME'} || $script_name),
+        DOCUMENT_ROOT   =>  ($args{'DOCUMENT_ROOT'} || $ENV{'DOCUMENT_ROOT'} || cwd()),
         DATE_GMT        =>  $gmt,
         DATE_LOCAL      =>  $loc,
-        LAST_MODIFIED   =>  $lmod,
-        DOCUMENT_NAME   =>  ($args{'DOCUMENT_NAME'} || $script_name),
-        DOCUMENT_ROOT   =>  ($args{'DOCUMENT_ROOT'} || $ENV{DOCUMENT_ROOT}),
                                 };
 
+    $self->{_timefmt_default} = "%A, %d-%B-%Y %T %Z"; # APACHE DEFAULT https://httpd.apache.org/docs/2.2/mod/mod_include.html#ssitimeformat
+    
     $self->{'_config'}        = {  # NOTE: TODO: get these from apache config
         errmsg  =>  ($args{'errmsg'}  || '[an error occurred while processing this directive]'),
         sizefmt =>  ($args{'sizefmt'} || 'abbrev'),
-        timefmt =>  ($args{'timefmt'} ||  undef),
+        timefmt =>  ($args{'timefmt'} ||  $self->{_timefmt_default}),
         SSIUndefinedEcho =>  ($args{'SSIUndefinedEcho'} ||  '(none)'),
-        _verbose_errors  =>  ($args{'_verbose_errors'}  ||  0)  # NOTE: TODO: document this option
+        _enable_exec_cmd  =>  ($args{'_enable_exec_cmd'}  ||  0),
+        _verbose_errors  =>  ($args{'_verbose_errors'}  ||  0)
                                 };
+                                
+    
+    $self->{'_variables'}->{LAST_MODIFIED}   =  $lmod; # needs to be specified after the above, since it requires DOCUMENT_ROOT to be populated
 
     $self->{_max_recursions} = $args{MAX_RECURSIONS} || 100; # no "infinite" loops
     $self->{_recursions} = {};
@@ -121,6 +139,11 @@ sub new {
     return $self;
 }
 
+sub _enable_exec_cmd {
+    my $self = shift;
+    $self->{'_config'}->{'_enable_exec_cmd'} = $_[0];
+}
+
 sub TIEHANDLE {
     my($class,%args) = @_;
     my $self = $class->new(%args);
@@ -375,7 +398,7 @@ sub echo {
     if ($key eq 'encoding') {
         $encoding = $var;		 # NOTE: TODO: handle encoding.
         ($key,$var) = @_[3,4];
-        $var = $key if (!$var);
+        $var = $key if (!defined($var));
     }
     
     if($var eq 'DATE_LOCAL') {
@@ -398,6 +421,9 @@ sub printenv {
 sub include {
 	$DEBUG and do { local $" = "','"; warn "DEBUG: include('@_')\n" };
     my($self,$type,$filename) = @_;
+    
+    $self->_init_CWD;
+
     if(lc $type eq 'file') {
           return $self->_include_file($filename);
     } elsif(lc $type eq 'virtual') {
@@ -406,87 +432,77 @@ sub include {
           return $self->SSI_ERROR_FLUSH("arg to include is '$type'. It must be one of: 'file' or 'virtual'.");
     }
 }
-
-sub _include_file {
-    $DEBUG and do { local $" = "','"; warn "DEBUG: _include_file('@_')\n" };
+sub _load_file {
     my($self,$filename) = @_;
-
-        # get the filename to open
-    $filename = catfile($FindBin::Bin,$filename) unless -e $filename;
-
-    # if we've reached MAX_RECURSIONS for this filename, warn and return the error
-    if(++$self->{_recursions}->{$filename} >= $self->{_max_recursions}) {
-            return $self->SSI_ERROR_FLUSH("the maximum number of 'include file' recursions has been exceeded for '$filename'.");
-    }
-
+    
+    unless (-e $filename) {
+        return $self->SSI_ERROR("File does not exist: $filename");
+        }
+    
         # open the file, or warn and return an error
     my $fh = do { local *STDIN };
     open($fh,$filename) or do {
         return $self->SSI_ERROR_FLUSH("failed to open file ($filename): $!");
     };
-
-        # process the included file and return the result
-    return $self->process(join '',<$fh>);
+    return join '',<$fh>;
 }
 
+#   http://httpd.apache.org/docs/2.2/howto/ssi.html
+# The *file* attribute is a file path, relative to the current directory.
+# That means that it cannot be an absolute file path (starting with /), nor can it contain ../ as part of that path.
+
+sub _include_file {
+    $DEBUG and do { local $" = "','"; warn "DEBUG: _include_file('@_')\n" };
+    my($self,$filename) = @_;
+
+    if ($filename =~ m{(^/|\.\.)}) {
+        return $self->SSI_ERROR_FLUSH("unable to include '$filename' in parsed file"); 
+        }
+    my $filepath = catfile($self->{_CWD},$filename);
+    return $self->_load_file($filepath) ||  $self->SSI_ERROR_FLUSH("unable to include \"$filename\" in parsed file ");
+}
+
+#
+#   http://httpd.apache.org/docs/2.2/howto/ssi.html
+# The *virtual* attribute is probably more useful, and should specify a URL relative to the _document being served_.
+# It can start with a /, but _must be on the same server_ as the file being served.
+
+		# NOTE: This could be greatly affected by mod_rewrite
+		
 sub _include_virtual {
     $DEBUG and do { local $" = "','"; warn "DEBUG: _include_virtual('@_')\n" };
     my($self,$filename) = @_;
+    my $file;
 
-    # if this is a local file that we can just read, let's do that instead of getting it virtually
     if($filename =~ m|^/(.+)|) { # could be on the local server: absolute filename, relative to ., relative to $ENV{DOCUMENT_ROOT}
-        my $file = $1;
-        if(-e '/'.$file) { # back to the original
-                $file = '/'.$file;
-        } elsif(-e catfile($self->{'_variables'}->{'DOCUMENT_ROOT'},$file)) {
-                $file = catfile($self->{'_variables'}->{'DOCUMENT_ROOT'},$file);
-        } elsif(-e catfile($FindBin::Bin,$file)) {
-                # $file = atfile($FindBin::Bin,$file);		 # <----- NOTE: is this a typo here??
-                $file = catfile($FindBin::Bin,$file);		 # fixing it just in case
+       $file = catfile($self->{'_variables'}->{'DOCUMENT_ROOT'},   $1);
+    }
+    else {
+       $file = catfile($self->{_CWD},   $filename);
+    }
+
+        # if we've reached MAX_RECURSIONS for this filename, warn and return the error
+    if(++$self->{_recursions}->{$file} >= $self->{_max_recursions}) {
+        return $self->SSI_ERROR_FLUSH("the maximum number of 'include virtual' recursions has been exceeded for '$filename'.");
+    }
+            
+    if (-e $file) {
+        my $dir = ( File::Spec->splitpath( $file ) )[1];
+        local $self->{_CWD} = $dir; # set the current working directory, because subsequent includes are relative to that..
+                
+        # process the included file and return the result
+        if (my $file_conts = $self->_load_file($file)) {
+            return $self->process($file_conts);
+        }
+        else {
+            return $self->SSI_ERROR_FLUSH("unable to include \"$filename\" in parsed file ");
         }
-        return $self->_include_file($file) if -e $file;
     }
-
-        # create the URI to get(), or warn and return the error
-    my $uri = eval {
-        my $uri = URI->new($filename);
-        $uri->scheme($uri->scheme || ($ENV{HTTPS} ? 'https' : 'http')); # ??
-        $uri->host($uri->host || $ENV{'HTTP_HOST'} || $ENV{'SERVER_NAME'} || 'localhost');
-        $uri;
-    } or do {
-        return $self->SSI_ERROR_FLUSH("failed to create a URI based on '$filename'.");
-    };
-    if($@) {
-        return $self->SSI_ERROR_FLUSH("failed to create a URI based on '$filename'.");
+    else {
+         return $self->SSI_ERROR_FLUSH("File does not exist: $file"."\nunable to include \"$filename\" in parsed file "); 
     }
-
-    # get the content of the request
-    $self->{_ua} ||= $self->_get_ua();
-    my $url = $uri->canonical;
-
-    # have we reached MAX_RECURSIONS?
-    if(++$self->{_recursions}->{$url} >= $self->{_max_recursions}) {
-        return $self->SSI_ERROR_FLUSH("the maximum number of 'include virtual' recursions has been exceeded for '$url'.");
-    }
-
-    my $response = $self->{_ua}->get($url);
-
-    # is it a success?
-    unless($response->is_success) {
-        return $self->SSI_ERROR_FLUSH("failed to get('$url'): ".$response->status_line.".");
-    }
-        # process the included content and return the result
-    return $self->process($response->content);
 }
 
-sub _get_ua {
-    my $self = shift;
-    my %conf = ();
-    $conf{agent} = $ENV{HTTP_USER_AGENT} if $ENV{HTTP_USER_AGENT};
-    my $ua = LWP::UserAgent->new(%conf);
-    $ua->cookie_jar($self->{_cookie_jar});
-    return $ua;
-}
 
 sub cookie_jar {
     my $self = shift;
@@ -509,114 +525,110 @@ sub exec {
 
 sub _exec_cmd {
     my($self,$filename) = @_;
+    
+    unless ($self->{'_config'}->{'_enable_exec_cmd'}) {
+        return $self->SSI_ERROR_FLUSH("directive 'exec cmd' is disabled. Set '_enable_exec_cmd' to 1 to enable. ");
+    }
 
     # have we reached MAX_RECURSIONS?
-    if(++$self->{_recursions}->{$filename} >= $self->{_max_recursions}) {
+    if (++$self->{_recursions}->{$filename} >= $self->{_max_recursions}) {
         return $self->SSI_ERROR_FLUSH("the maximum number of 'exec cmd' recursions has been exceeded for '$filename'.");
     }
 
-    my $output = `$filename`; # security here is mighty bad.
+    my $output = `$filename`; # NOTE: security here is mighty bad.
 
     # was the command a success?
     if($?) {
-        return $self->SSI_ERROR_FLUSH("`$filename` was not successful.");
+        return $self->SSI_ERROR_FLUSH("exec cmd of `$filename` was not successful."); # NOTE: FIXME msg
     }
 
-        # process the output, and return the result
-    return $self->process($output);
+    return $output;
 }
 
-sub _exec_cgi { # no relative $filename allowed.
+sub _exec_cgi {
     my($self,$filename) = @_;
-
-    # have we reached MAX_RECURSIONS?
-    if(++$self->{_recursions}->{$filename} >= $self->{_max_recursions}) {
-        return $self->SSI_ERROR_FLUSH("the maximum number of 'exec cgi' recursions has been exceeded for '$filename'.");
-    }
-
-    # create the URI from the filename
-    my $uri = eval {
-        my $uri = URI->new($filename);
-        $uri->scheme($uri->scheme || ($ENV{HTTPS} ? 'https' : 'http')); # ??
-        $uri->host($uri->host || $ENV{'HTTP_HOST'} || $ENV{'SERVER_NAME'});
-        $uri->query($uri->query || $ENV{'QUERY_STRING'});
-        $uri;
-    } or do {
-        return $self->SSI_ERROR_FLUSH("failed to create a URI from '$filename'.");
-    };
-    if($@) {
-        return $self->SSI_ERROR_FLUSH("failed to create a URI from '$filename'.");
-    }
-
-    # get the content
-    $self->{_ua} ||= $self->_get_ua();
-    my $url = $uri->canonical;
-    my $response = $self->{_ua}->get($url);
-
-    # success?
-    unless($response->is_success) {
-          return $self->SSI_ERROR_FLUSH("failed to get('$filename').");
-    }
-
-        # process the content and return the result
-    return $self->process($response->content);
+    return $self->_include_virtual($filename);  # NOTE: FIXME -- exec cgi is simply an alias to include virtual!!
 }
 
 sub flastmod {
     my($self,$type,$filename) = @_;
 
+    $self->_init_CWD;
+    
     if(lc $type eq 'file') {
-            $filename = catfile($FindBin::Bin,$filename) unless -e $filename;
+        $filename = catfile($self->{_CWD},$filename) unless -e $filename;
     } elsif(lc $type eq 'virtual') {
-            $filename = catfile($self->{'_variables'}->{'DOCUMENT_ROOT'},$filename)
-            unless $filename =~ /$self->{'_variables'}->{'DOCUMENT_ROOT'}/;
+        $filename = catfile($self->{'_variables'}->{'DOCUMENT_ROOT'},$filename)
+        unless $filename =~ /$self->{'_variables'}->{'DOCUMENT_ROOT'}/;
     } else {
-            return $self->SSI_ERROR_FLUSH("the first argument to flastmod is '$type'. It must be one of: 'file' or 'virtual'.");
+        return $self->SSI_ERROR_FLUSH("the first argument to flastmod is '$type'. It must be one of: 'file' or 'virtual'.");
     }
     
     unless(-e $filename) {
-            return $self->SSI_ERROR_FLUSH("flastmod failed to find '$filename'.");
-            }
+        return $self->SSI_ERROR_FLUSH("flastmod failed to find '$filename'.");
+        }
 
     my $flastmod = (stat $filename)[9];
 
-    if($self->{'_config'}->{'timefmt'}) {
-            my @localtime = localtime($flastmod); # need this??
-            return Date::Format::strftime($self->{'_config'}->{'timefmt'},@localtime);
-    } else {
-            return scalar localtime($flastmod);
-    }
+    my @localtime = localtime($flastmod); # need this??
+    return Date::Format::strftime($self->{'_config'}->{'timefmt'} || $self->{_timefmt_default} , @localtime); 
 }
 
 sub fsize {
     my($self,$type,$filename) = @_;
 
+    $self->_init_CWD;
+    
     if(lc $type eq 'file') {
-            $filename = catfile($FindBin::Bin,$filename) unless -e $filename;
+        $filename = catfile($self->{_CWD},$filename) unless -e $filename;
     } elsif(lc $type eq 'virtual') {
-            $filename = catfile($ENV{'DOCUMENT_ROOT'},$filename) unless $filename =~ /$ENV{'DOCUMENT_ROOT'}/;
+        $filename = catfile($self->{'_variables'}->{'DOCUMENT_ROOT'},$filename) unless $filename =~ /$self->{'_variables'}->{'DOCUMENT_ROOT'}/;
     } else {
-            return $self->SSI_ERROR_FLUSH("the first argument to fsize is '$type'. It must be one of: 'file' or 'virtual'.");
+        return $self->SSI_ERROR_FLUSH("the first argument to fsize is '$type'. It must be one of: 'file' or 'virtual'.");
     }
     unless(-e $filename) {
-            return $self->SSI_ERROR_FLUSH("fsize failed to find '$filename'.");
+        return $self->SSI_ERROR_FLUSH("fsize failed to find '$filename'.");
     }
         
     my $fsize = (stat $filename)[7];
     
     if(lc $self->{'_config'}->{'sizefmt'} eq 'bytes') {
-            1 while $fsize =~ s/^(\d+)(\d{3})/$1,$2/g;
-            return $fsize;
+        1 while $fsize =~ s/^(\d+)(\d{3})/$1,$2/g; # add commas to the 10^3 position markers
+        return $fsize;
     } else { # abbrev
-            # gratefully lifted from Apache::SSI
-            return "   0k" unless $fsize;
-            return "   1k" if $fsize < 1024;
-            return sprintf("%4dk", ($fsize + 512)/1024) if $fsize < 1048576;
-            return sprintf("%4.1fM", $fsize/1048576.0) if $fsize < 103809024;
-            return sprintf("%4dM", ($fsize + 524288)/1048576) if $fsize < 1048576;
+        # ---------------
+        # adapted from the horrific code in apache itself            
+        # http://svn.apache.org/viewvc/apr/apr/trunk/strings/apr_strings.c  # apr_strfsize
+            
+        return sprintf("%3d",$fsize)  if ($fsize < 973); # bytes
+        my $remain;
+        foreach my $units (qw(K M G T P E)) {
+            $remain = $fsize % 1024;
+            $fsize >>= 10; # /1024 ie, 2^10
+            next if ($fsize >= 973);
+            if ($fsize < 9 || ($fsize == 9 && $remain < 973)) { # < 9973?
+                if ( ($remain = (($remain * 5) + 256) / 512) >= 10)
+                    { return sprintf("%d.0%s", ++$fsize, $units); }
+                return sprintf("%d.%d%s", $fsize, $remain, $units);
+                }
+            if ($remain >= 512) # >= 10512?
+                { $fsize++; } # round up?
+            return sprintf("%3d%s", $fsize, $units);
+            }
     }
 }
 
+
+sub _init_CWD {
+    my($self) = @_;
+    my $req = $ENV{REQUEST_URI} || '';
+    my $dr = $self->{'_variables'}->{'DOCUMENT_ROOT'};
+    unless (defined($self->{_CWD}))
+        { $self->{_CWD} = ( File::Spec->splitpath( $dr . $req) )[1] ;}
+}
+
+
+
 #
 # if/elsif/else/endif and related methods
 #
@@ -664,7 +676,7 @@ sub _test {
             # if op, and no RHS or no LHS, FAIL
             # if op, do op.. return [pound != test(LHS)] op [test(LHS)]
             if ($OP) {
-                # LOGICAL COMPARISON && and ||
+                 # LOGICAL COMPARISON && and ||
                     # NOTE:  && and || have equal precedence
 
                 if ($LHS=~/^\s*$/) {
@@ -918,7 +930,7 @@ sub else {
 sub endif {
     my $self = shift;
     if (! $self->_in_if() ) {
-        $self->SSI_WARN("Incorrect use of endif ssi directive: no preceeding 'if'.");
+        $self->SSI_WARN("Incorrect use of endif ssi directive: no preceeding 'if'."); # NOTE: just a "warn"
         }
     else
         { $self->_leaving_if(); }
@@ -1077,7 +1089,8 @@ the parsing engine and test expression code, which is now no longer "perlish". A
 feature could be added to do perlish expressions via something like a "perl_expr" directive,
 but since I didn't need it, I didn't implement it. The rest is basically the same,
 so a lot of the documentation below is taken directly from CGI::SSI. 
-Needless to say: "USE AT YOUR OWN RISK".
+
+Needless to say: B<"USE AT YOUR OWN RISK">.
 
 CGI::apacheSSI is meant to be an easy way to add the ability to filter and parse (even nested!)
 existing shtml in CGI scripts without the need to modify any of the files to be parsed.
@@ -1096,6 +1109,45 @@ Be aware that Apache's C<mod_include> treats single quotes ' slightly differentl
 double quotes " and backticks ` when you use those to wrap your expressions, and therefore
 so will CGI::apacheSSI.
 
+I'm sure there are some interesting applications of this module so please do let me know if you
+use it for anything or if you cannibalized any of the code for a different project.
+
+=head1 UNIMPLEMENTED FEATURES AND IMPORTANT INFO
+
+There are many features not currently implemented, but these are some of the ones I am aware of
+that will definitely impede your ability to use this module as a faithful emulator of
+C<mod_include>.
+
+=head2 All server configuration directives are ignored
+
+That means things like C<Options IncludesNOEXEC> are not taken into consideration, and you might
+end up parsing pages you might not otherwise be able to.
+
+=head2 C<include virtual> for cgi scripts and mod_rewrite urls
+
+The C<include virtual> call cannot currently handle calls to cgi scripts or mod_rewrite urls.
+It is completely file-based, and is not currently capable of doing apache internal calls.
+
+=head2 C<exec cgi> is simply an alias to C<include virtual>
+
+In general, you are always advised to avoid the C<exec> directive, and C<apache::SSI> implements
+C<exec cgi> as an alias for C<include virtual>.
+
+=head2 C<exec cmd> needs to be explicitly enabled
+
+Given the above, C<exec cmd> is disabled by default, since, when enabled, it is simply a sub-call using
+backticks ``, leading to a big security liability. This can be set as follows:
+
+    my $ssi = CGI::apacheSSI->new('_enable_exec_cmd' => 1);   # at instantiation time
+    
+    # or during usage: 
+    $ssi->_enable_exec_cmd(1);    # to enable
+    $ssi->_enable_exec_cmd(0);    # to disable
+
+=head2 C<echo> encoding
+
+The C<echo> command silently ignores "encoding" specifications.
+
 =head1 USAGE
 
 Most of the time, you'll simply want to filter shtml through STDOUT 
